import logging
from time import sleep

from nuclei.client import NucleiClient
from requests import Response


def wait_until_ticket_is_ready(client: NucleiClient, ticket: Response) -> None:
    if ticket.status_code != 200:
        raise RuntimeError(rf"{ticket.text}")

    status = "STARTED"
    sleep_time = 0.05
    while status in ["PENDING", "STARTED", "RETRY"]:
        sleep_time = min(sleep_time * 2, 10)
        sleep(sleep_time)
        status_response = client.call_endpoint(
            "VibraCore", "/get-task-status", schema=ticket.json(), return_response=True
        )

        # Check if the status response is OK
        if status_response.status_code != 200:
            raise RuntimeError(rf"{status_response.text}")

        status = status_response.json()["state"]

    # If the status is FAILURE, raise an error
    if status == "FAILURE":
        # Get the task-status failure message
        failure_message = status_response.json()["msg"]

        # Try to get the task-result failure message
        try:
            result_response = client.call_endpoint(
                "VibraCore",
                "/get-task-results",
                schema=ticket.json(),
                return_response=True,
            )
            failure_message = result_response.text

        # Raise the obtained failure message
        finally:
            raise RuntimeError(failure_message)


def get_impact_force_report(client: NucleiClient, payload: dict) -> bytes:
    """
    Wrapper around the VibraCore endpoint "/impact-force/report".

    Parameters
    ----------
    client: NucleiClient
        client object created by [nuclei](https://github.com/cemsbv/nuclei)
    payload: dict
        the payload of the request, can be created by calling `create_multi_cpt_impact_force_report_payload()`
    """
    logging.info(
        "Generate report... \n"
        "Depending on the amount of CPT's this can take a while."
    )
    ticket = client.call_endpoint(
        "VibraCore",
        "/impact-force/report",
        schema=payload,
        return_response=True,
    )

    wait_until_ticket_is_ready(client=client, ticket=ticket)

    return client.call_endpoint("VibraCore", "/get-task-results", schema=ticket.json())


def get_impact_force_calculation(client: NucleiClient, payload: dict) -> bytes:
    """
    Wrapper around the VibraCore endpoint "/impact-force/calculation/multi".

    Parameters
    ----------
    client: NucleiClient
        client object created by [nuclei](https://github.com/cemsbv/nuclei)
    payload: dict
        the payload of the request, can be created by calling `create_multi_cpt_impact_force_payload()`
    """
    logging.info(
        "Calculation impact force... \n"
        "Depending on the amount of CPT's this can take a while."
    )
    ticket = client.call_endpoint(
        "VibraCore",
        "/impact-force/calculation/multi",
        schema=payload,
        return_response=True,
    )

    wait_until_ticket_is_ready(client=client, ticket=ticket)

    return client.call_endpoint("VibraCore", "/get-task-results", schema=ticket.json())


def get_prepal_calculation(client: NucleiClient, payload: dict) -> bytes:
    """
    Wrapper around the VibraCore endpoint "/prepal/validation/multi".

    Parameters
    ----------
    client: NucleiClient
        client object created by [nuclei](https://github.com/cemsbv/nuclei)
    payload: dict
        the payload of the request, can be created by calling `create_prepal_payload()`
    """
    logging.info(
        "Prepal prediction... \n"
        "Depending on the amount of buildings this can take a while."
    )
    ticket = client.call_endpoint(
        "VibraCore",
        "/prepal/validation/multi",
        schema=payload,
        return_response=True,
    )

    wait_until_ticket_is_ready(client=client, ticket=ticket)

    return client.call_endpoint("VibraCore", "/get-task-results", schema=ticket.json())


def get_cur166_calculation(client: NucleiClient, payload: dict) -> bytes:
    """
    Wrapper around the VibraCore endpoint "/cur166/validation/multi".

    Parameters
    ----------
    client: NucleiClient
        client object created by [nuclei](https://github.com/cemsbv/nuclei)
    payload: dict
        the payload of the request, can be created by calling `create_cur166_payload()`
    """
    logging.info(
        "CUR166 prediction... \n"
        "Depending on the amount of buildings this can take a while."
    )
    ticket = client.call_endpoint(
        "VibraCore",
        "/cur166/validation/multi",
        schema=payload,
        return_response=True,
    )

    wait_until_ticket_is_ready(client=client, ticket=ticket)

    return client.call_endpoint("VibraCore", "/get-task-results", schema=ticket.json())


def get_cur166_report(client: NucleiClient, payload: dict) -> bytes:
    """
    Wrapper around the VibraCore endpoint "/cur166/report".

    Parameters
    ----------
    client: NucleiClient
        client object created by [nuclei](https://github.com/cemsbv/nuclei)
    payload: dict
        the payload of the request, can be created by calling `create_vibration_report_payload()`
    """
    logging.info(
        "Generate report... \n"
        "Depending on the amount of buildings this can take a while."
    )
    ticket = client.call_endpoint(
        "VibraCore",
        "/cur166/report",
        schema=payload,
        return_response=True,
    )

    wait_until_ticket_is_ready(client=client, ticket=ticket)

    return client.call_endpoint("VibraCore", "/get-task-results", schema=ticket.json())


def get_prepal_report(client: NucleiClient, payload: dict) -> bytes:
    """
    Wrapper around the VibraCore endpoint "/cur166/report".

    Parameters
    ----------
    client: NucleiClient
        client object created by [nuclei](https://github.com/cemsbv/nuclei)
    payload: dict
        the payload of the request, can be created by calling `create_vibration_report_payload()`
    """
    logging.info(
        "Generate report... \n"
        "Depending on the amount of buildings this can take a while."
    )
    ticket = client.call_endpoint(
        "VibraCore",
        "/prepal/report",
        schema=payload,
        return_response=True,
    )

    wait_until_ticket_is_ready(client=client, ticket=ticket)

    return client.call_endpoint("VibraCore", "/get-task-results", schema=ticket.json())



from __future__ import annotations

import json
from dataclasses import dataclass
from typing import Any, Sequence, Tuple

import geopandas as gpd
import matplotlib.pyplot as plt
import pandas as pd
from mpl_toolkits.axes_grid1.anchored_artists import AnchoredSizeBar


@dataclass(frozen=True)
class MultiCalculationData:
    """
    Dataclass that holds the information from `/impact-force/multi`

    Attributes
    -----------
    gdf: gpd.GeoDataFrame

    """

    gdf: gpd.GeoDataFrame

    @classmethod
    def from_api_response(cls, response_dict: dict) -> "MultiCalculationData":
        """
        Stores the response of the VibraCore endpoint

        Parameters
        ----------
        response_dict:
           The resulting response of a call to `/impact-force/multi`
        """
        return cls(
            gdf=gpd.read_file(json.dumps(response_dict), driver="GeoJSON"),
        )

    def plot(
        self,
        figsize: Tuple[float, float] = (10.0, 12.0),
        settings: dict | None = None,
        **kwargs: Any,
    ) -> plt.Figure:
        """
        Plots the CPT and soil table data.

        Parameters
        ----------
        figsize:
            Size of the activate figure, as the `plt.figure()` argument.
        settings:
            Plot settings used in plot: default settings are:

            .. code-block:: python

                {
                    "hue": "max",
                    "title": "Maximale slagkracht [kN]",
                    "xlabel": "X-coördinaat",
                    "ylabel": "Y-coördinaat"
                }
        **kwargs:
            All additional keyword arguments are passed to the `pyplot.subplots()` call.

        Returns
        -------
        fig:
            The matplotlib Figure
        """
        kwargs_settings = {
            "hue": "max",
            "title": "Maximale slagkracht [kN]",
            "xlabel": "X-coördinaat",
            "ylabel": "Y-coördinaat",
        }

        if settings:
            kwargs_settings.update(settings)

        if kwargs_settings["hue"] not in ["max", "base", "Q90", "Q95"]:
            raise ValueError(
                f"The value {kwargs_settings['hue']} is not in [max, base, Q90, Q95]"
            )

        kwargs_subplot = {
            "figsize": figsize,
            "tight_layout": True,
        }

        kwargs_subplot.update(kwargs)

        fig, axes = plt.subplots(**kwargs_subplot)

        # Create figure
        self.gdf.plot(
            kwargs_settings.get("hue", "max"),
            ax=axes,
            legend=True,
            figsize=(15, 10),
            legend_kwds={"orientation": "vertical"},
            cmap="RdYlGn_r",  # 'jet',
            markersize=150,
            marker="v",
            aspect=1,
        )

        # add CPT name to plot
        for idx, row in self.gdf.iterrows():
            axes.annotate(
                row["id"],
                xy=(row["x"], row["y"]),
                horizontalalignment="center",
                fontsize=20,
                xytext=(row["x"], row["y"] + 3),
            )

        # Set label and title for figure
        axes.set_xlabel(xlabel=kwargs_settings.get("xlabel", ""), size=15)
        axes.set_ylabel(ylabel=kwargs_settings.get("ylabel", ""), size=15)
        axes.set_title(label=kwargs_settings.get("title", ""), size=25)

        # Add north arrow
        x, y, arrow_length = 0.95, 0.98, 0.1
        axes.annotate(
            "N",
            xy=(x, y),
            xytext=(x, y - arrow_length),
            arrowprops=dict(facecolor="black", width=5, headwidth=15),
            ha="center",
            va="center",
            fontsize=20,
            xycoords=axes.transAxes,
        )

        # scale bar
        scalebar = AnchoredSizeBar(
            axes.transData,
            20,
            "20 m",
            "lower left",
            pad=1,
            color="black",
            frameon=True,
            size_vertical=2,
        )
        axes.add_artist(scalebar)

        return fig


@dataclass(frozen=True)
class ImpactForceTable:
    """
    Object that contains the impact force related data-traces.

    Attributes:
    ------------
    betaFrictionalResistance: Sequence[float]
        beta frictional resistance [kN]
    betaPointResistance: Sequence[float]
        beta point resistance [kN]
    correctedConeResistance: Sequence[float]
        corrected cone resistance [Mpa]
    depthOffset: Sequence[float]
        CPT depth [m w.r.t. Reference]
    frictionalResistance: Sequence[float]
        frictional resistance [kN]
    pointResistance: Sequence[float]
        point resistance [kN]
    sheetResistance: Sequence[float]
        sheet resistance [kN]
    slotResistance: Sequence[float]
        slot resistance [kN]
    totalResistance: Sequence[float]
        total resistance [kN]
    """

    betaFrictionalResistance: Sequence[float] | None
    betaPointResistance: Sequence[float] | None
    correctedConeResistance: Sequence[float] | None
    depthOffset: Sequence[float] | None
    frictionalResistance: Sequence[float] | None
    pointResistance: Sequence[float] | None
    sheetResistance: Sequence[float] | None
    slotResistance: Sequence[float] | None
    totalResistance: Sequence[float] | None

    def __post_init__(self) -> None:
        raw_lengths = []
        for values in self.__dict__.values():
            if values:
                raw_lengths.append(len(values))
        if len(list(set(raw_lengths))) > 1:
            raise ValueError("All values in this dataclass must have the same length.")

    @classmethod
    def from_api_response(cls, response_dict: dict) -> "ImpactForceTable":
        """
        Stores the response of the VibraCore endpoint

        Parameters
        ----------
        response_dict:
           The resulting response of a call to `impact-force/calculation/single`
        """
        return cls(
            betaFrictionalResistance=response_dict.get("betaFrictionalResistance"),
            betaPointResistance=response_dict.get("betaPointResistance"),
            correctedConeResistance=response_dict.get("correctedConeResistance"),
            depthOffset=response_dict.get("depthOffset"),
            frictionalResistance=response_dict.get("frictionalResistance"),
            pointResistance=response_dict.get("pointResistance"),
            sheetResistance=response_dict.get("sheetResistance"),
            slotResistance=response_dict.get("slotResistance"),
            totalResistance=response_dict.get("totalResistance"),
        )

    @property
    def dataframe(self) -> pd.DataFrame:
        """The pandas.DataFrame representation"""
        return pd.DataFrame(self.__dict__).dropna(axis="rows", how="any")  # type: ignore


@dataclass(frozen=True)
class SingleCalculationData:
    """
    Object that contains the impact force related data.

    Attributes:
    ------------
    table: ImpactForceTable
        table object
    installationLevel: float
        installation level [m w.r.t. Reference]
    maximumForce: float
        total resistance [kN]
    pointForce: float
        resistance at the base [kN]
    """

    table: ImpactForceTable
    installationLevel: float
    maximumForce: float
    pointForce: float | None

    @classmethod
    def from_api_response(cls, response_dict: dict) -> "SingleCalculationData":
        """
        Stores the response of the VibraCore endpoint

        Parameters
        ----------
        response_dict:
           The resulting response of a call to `impact-force/calculation/single`
        """
        return cls(
            table=ImpactForceTable.from_api_response(response_dict["data"]),
            installationLevel=response_dict["installationLevel"],
            maximumForce=response_dict["maximumForce"],
            pointForce=response_dict.get("pointForce"),
        )

    def plot(
        self,
        figsize: Tuple[float, float] = (10.0, 12.0),
        **kwargs: Any,
    ) -> plt.Figure:
        """
        Plots the resistance data.

        Parameters
        ----------
        figsize:
            Size of the activate figure, as the `plt.figure()` argument.
        **kwargs:
            All additional keyword arguments are passed to the `pyplot.subplots()` call.

        Returns
        -------
        fig:
            The matplotlib Figure
        """
        kwargs_subplot = {
            "figsize": figsize,
            "tight_layout": True,
        }

        kwargs_subplot.update(kwargs)

        fig, axes = plt.subplots(**kwargs_subplot)

        for item in [
            "totalResistance",
            "frictionalResistance",
            "slotResistance",
            "pointResistance",
        ]:
            axes.plot(
                self.table.__getattribute__(item),
                self.table.depthOffset,
                label=item,
            )

        axes.axvline(
            x=self.installationLevel,
            c="k",
            linestyle="dashed",
            label=f"max force ({self.maximumForce.__round__(2)})",
        )
        axes.axhline(
            y=self.installationLevel,
            c="k",
            linestyle="dotted",
            label=f"installation level ({self.installationLevel})",
        )
        axes.set_xlabel("Resistance [kN]")
        axes.grid()

        return fig



from __future__ import annotations

from copy import deepcopy
from typing import Any, List, Literal, Sequence, Tuple

import geopandas as gpd
import matplotlib.patches as patches
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from numpy.typing import NDArray
from scipy.interpolate import interp1d
from shapely.geometry import LineString, Point, Polygon

from pyvibracore.results.plot_utils import _north_arrow, _scalebar

# CUR 166-1997 Tabel 5.20 Factor Cfc
CFC_FACTOR_FLOORS = {
    "driving": {
        "concrete": {"Cfc": 1.4, "Vfc": 0.17},
        "wood": {"Cfc": 1.4, "Vfc": 0.17},
    },
    "vibrate": {
        "concrete": {"Cfc": 1.7, "Vfc": 0.27},
        "wood": {"Cfc": 2.5, "Vfc": 0.44},
    },
}

# SBR-B Trillingen meet en beoordelingsrichtlijnen hinder voor personen in gebouwen [2002] art. 10.5.4 Table 4
TARGET_VALUE = {
    "<= 1 day": [0.8, 6, 0.4],
    "2 days": [0.72, 6, 0.38],
    "3 days": [0.64, 6, 0.34],
    "4 days": [0.56, 6, 0.36],
    "5 days": [0.48, 6, 0.32],
    ">= 6 days; <26 days": [0.4, 6, 0.3],
    ">= 26 days; <78 days": [0.3, 6, 0.2],
    "Unlimited": {
        "short-term": {
            "woonfunctie": [0.2, 0.8, 0.1],
            "gezondheidsfunctie": [0.2, 0.8, 0.1],
            "onderwijsfunctie": [0.3, 1.2, 0.15],
        },
        "repeated-short-term": {
            "woonfunctie": [0.2, 0.8, 0.1],
            "gezondheidsfunctie": [0.2, 0.8, 0.1],
            "onderwijsfunctie": [0.3, 1.2, 0.15],
        },
        "continuous": {
            "woonfunctie": [0.1, 0.4, 0.05],
            "gezondheidsfunctie": [0.1, 0.4, 0.05],
            "onderwijsfunctie": [0.15, 0.6, 0.07],
        },
    },
}


def _get_target_value(
    vibration_type: Literal["short-term", "repeated-short-term", "continuous"],
    building_function: str,
) -> dict:
    if "woonfunctie" in building_function:
        _building_function = "woonfunctie"
    elif "gezondheidsfunctie" in building_function:
        _building_function = "gezondheidsfunctie"
    elif "onderwijsfunctie" in building_function:
        _building_function = "onderwijsfunctie"
    else:
        _building_function = "other"

    _body = deepcopy(TARGET_VALUE)
    _body["Unlimited"] = TARGET_VALUE["Unlimited"][vibration_type].get(  # type: ignore
        _building_function, [np.nan, np.nan, np.nan]
    )
    return _body


def _nuisance_prediction(
    target_value_one: List[float] | NDArray | Sequence,
    target_value_two: List[float] | NDArray | Sequence,
    target_value_three: List[float] | NDArray | Sequence,
    vibration_velocity_eff: List[float] | NDArray,
    vibration_velocity_per: List[float] | NDArray,
    distance: List[float] | NDArray,
) -> NDArray:
    """
    Based on the 'SBR-B Trillingen meet en beoordelingsrichtlijnen hinder voor personen in gebouwen [2002]'.

    Parameters
    ----------
    target_value_one:
        target value SBR-B [2002] art. 10.5.1 [-]
    target_value_two:
        target value SBR-B [2002] art. 10.5.1 [-]
    target_value_three:
        target value SBR-B [2002] art. 10.5.1 [-]
    vibration_velocity_eff:
        vibration velocity [mm/s]
    vibration_velocity_per: list
        vibration velocity [mm/s]
    distance:
        distance with respect to building [m]

    Returns
    -------
    space: NDArray
    """

    df = pd.DataFrame(
        {
            "vibrationVelocity_per": vibration_velocity_per,
            "vibrationVelocity_eff": vibration_velocity_eff,
            "distance": distance,
        }
    ).drop_duplicates(subset=["vibrationVelocity_per", "vibrationVelocity_eff"])

    # interpolate and predict
    f_eff = interp1d(
        df["vibrationVelocity_eff"],
        df["distance"],
        kind="linear",
        assume_sorted=False,
        fill_value="extrapolate",
    )
    target_value_one_spaces = f_eff(target_value_one)
    target_value_two_spaces = f_eff(target_value_two)

    # interpolate and predict
    f_per = interp1d(
        df["vibrationVelocity_per"],
        df["distance"],
        kind="linear",
        assume_sorted=False,
        fill_value="extrapolate",
    )
    target_value_three_spaces = f_per(target_value_three)

    return np.min(
        [
            np.max([target_value_two_spaces, target_value_three_spaces], axis=0),
            target_value_one_spaces,
        ],
        axis=0,
    )


def df_nuisance(
    buildings: gpd.GeoDataFrame,
    response_dict: dict,
    building_name: str,
    vibration_type: Literal[
        "short-term", "repeated-short-term", "continuous"
    ] = "continuous",
    installation_type: Literal["vibrate", "driving"] = "vibrate",
    period: float = 10,
) -> pd.DataFrame:
    """
    Get a DataFrame that holds the distance of the different durations

    Parameters
    ----------
    buildings:
        GeoDataFrame of the input buildings
    response_dict:
        response of the single prepal or cur166 endpoint.
    building_name:
        name of the building
    vibration_type
        Based on the SBR A table 10.4.
    installation_type
        Based on CUR 166 3rd edition table 5.20 or 5.21
    period:
        Operating period of the building code [hours]
    period:
        Operating period of the building code [hours]

    Returns
    -------
    dataframe
    """
    building = buildings.get(buildings["name"] == building_name)
    if building.empty:
        raise ValueError(f"No buildings with name {building_name}.")

    levels = _get_target_value(
        vibration_type,
        building_function=building["gebruiksdoel"].item(),
    ).values()
    arr = np.array(response_dict["data"]["vibrationVelocity"]) / response_dict[
        "calculation"
    ].get("gamma", 1)
    a_one, a_two, a_three = [*zip(*levels)]

    # safety factors
    cfc = CFC_FACTOR_FLOORS[installation_type][building["material"].item()]["Cfc"]
    u_eff = 0.64 if vibration_type == "continuous" else 0.42

    distances = _nuisance_prediction(
        target_value_one=a_one,
        target_value_two=a_two,
        target_value_three=a_three,
        vibration_velocity_per=arr * cfc * u_eff * np.sqrt(period / 12),
        vibration_velocity_eff=arr * cfc * u_eff,
        distance=response_dict["data"]["distance"],
    )

    return pd.DataFrame(
        {
            building_name: distances,
        },
        index=list(TARGET_VALUE.keys()),
    )


def map_nuisance(
    buildings: gpd.GeoDataFrame,
    source_location: Point | LineString | Polygon,
    building_name: str,
    response_dict: dict,
    vibration_type: Literal[
        "short-term", "repeated-short-term", "continuous"
    ] = "continuous",
    installation_type: Literal["vibrate", "driving"] = "vibrate",
    period: float = 10,
    title: str = "Legend:",
    figsize: Tuple[float, float] = (10.0, 12.0),
    settings: dict | None = None,
    **kwargs: Any,
) -> plt.Figure:
    """
    Create map of the input building settings.

    Parameters
    ----------
    buildings:
        GeoDataFrame of the input buildings
    response_dict:
        response of the single prepal or cur166 endpoint.
    source_location:
        location of the vibration source
    building_name:
        name of the building
    vibration_type
        Based on the SBR A table 10.4.
    installation_type
        Based on CUR 166 3rd edition table 5.20 or 5.21
    period:
        Operating period of the building code [hours]
    title:
        Legend title
    figsize:
        Size of the activate figure, as the `plt.figure()` argument.
    settings:
        Plot settings used in plot: default settings are:

        .. code-block:: python

            {
                "source_location": {"label": "Trillingsbron", "color": "blue"},
                "levels": [
                    {
                        "label": "<= 1 day",
                        "color": "darkred",
                    },
                    {
                        "label": ">= 6 days; <26 days",
                        "color": "orange",
                    },
                    {
                        "label": ">= 26 days; <78 days",
                        "color": "green",
                    },
                ],
            }
    **kwargs:
        All additional keyword arguments are passed to the `pyplot.subplots()` call.

    Returns
    -------
    Figure
    """
    if settings is None:
        settings = {
            "source_location": {"label": "Trillingsbron", "color": "blue"},
            "levels": [
                {
                    "label": "<= 1 day",
                    "color": "darkred",
                },
                {
                    "label": ">= 6 days; <26 days",
                    "color": "orange",
                },
                {
                    "label": ">= 26 days; <78 days",
                    "color": "green",
                },
            ],
        }

    kwargs_subplot = {
        "figsize": figsize,
        "tight_layout": True,
    }

    kwargs_subplot.update(kwargs)

    fig, axes = plt.subplots(**kwargs_subplot)

    gpd.GeoSeries(source_location).plot(
        ax=axes, color=settings["source_location"]["color"], alpha=1, zorder=1, aspect=1
    )

    building = buildings.get(buildings["name"] == building_name)
    if building.empty:
        raise ValueError(f"No buildings with name {building_name}.")

    buildings.where(buildings["name"] == building_name).plot(
        ax=axes, zorder=2, color="gray", aspect=1
    )
    buildings.where(buildings["name"] != building_name).plot(
        ax=axes, zorder=2, color="lightgray", aspect=1
    )

    # plot contour
    levels = [
        _get_target_value(
            vibration_type,
            building_function=building["gebruiksdoel"].item(),
        )[values["label"]]
        for values in settings["levels"]
    ]
    arr = np.array(response_dict["data"]["vibrationVelocity"]) / response_dict[
        "calculation"
    ].get("gamma", 1)
    a_one, a_two, a_three = [*zip(*levels)]

    # safety factors
    cfc = CFC_FACTOR_FLOORS[installation_type][building["material"].item()]["Cfc"]
    u_eff = 0.64 if vibration_type == "continuous" else 0.42

    distances = _nuisance_prediction(
        target_value_one=a_one,
        target_value_two=a_two,
        target_value_three=a_three,
        vibration_velocity_per=arr * cfc * u_eff * np.sqrt(period / 12),
        vibration_velocity_eff=arr * cfc * u_eff,
        distance=response_dict["data"]["distance"],
    )
    colors = [values["color"] for values in settings["levels"]]
    for distance, color in zip(distances, colors):
        gpd.GeoSeries(building.buffer(distance).exterior).plot(
            ax=axes, zorder=3, color=color, aspect=1
        )

    # plot name
    for idx, row in buildings.iterrows():
        x = row.geometry.centroid.xy[0][0]
        y = row.geometry.centroid.xy[1][0]

        axes.annotate(
            idx,
            xy=(x, y),
            horizontalalignment="center",
        )

    # add legend
    axes.legend(
        title=title,
        title_fontsize=18,
        fontsize=15,
        loc="lower right",
        handles=[
            patches.Patch(
                facecolor=value["color"],
                label=value["label"],
                alpha=0.9,
                linewidth=2,
                edgecolor="black",
            )
            for value in settings["levels"]
        ],
    )

    _north_arrow(axes)
    _scalebar(axes)

    return fig



from __future__ import annotations

import json
import logging
import uuid
from copy import deepcopy
from typing import Literal

import geopandas as gpd
import numpy as np
import requests
from shapely.geometry import LineString, Point, Polygon, mapping

from .constants import SOIL_REFERENCE

BAG_WFS_URL = "https://service.pdok.nl/lv/bag/wfs/v2_0"
THRESHOLD = 10000


def get_buildings_geodataframe(
    west: float,
    south: float,
    east: float,
    north: float,
    category: Literal["one", "two"] = "two",
    monumental: bool = False,
    structural_condition: Literal["sensitive", "normal"] = "normal",
    vibration_sensitive: bool = False,
    thickness: float = 8.0,
    foundation_element: Literal[
        "shallow foundation", "concrete piles", "timber piles", "steel piles"
    ] = "shallow foundation",
    material_floor: Literal["concrete", "wood"] = "concrete",
    feature: Literal[
        "bag:pand",
        "bag:ligplaats",
        "bag:verblijfsobject",
        "bag:woonplaats",
        "bag:standplaats",
    ] = "bag:pand",
    pagesize: Literal["10", "20", "50", "100", "1000"] = "1000",
) -> gpd.GeoDataFrame:
    """
    Get a GeoDataFrame with the default values for CUR166 and PrePal methode.

    Parameters
    ----------
    west:
        west coordinate in rd new (EPSG:28992)
    south:
        south coordinate in rd new (EPSG:28992)
    east:
        east coordinate in rd new (EPSG:28992)
    north:
        north coordinate in rd new (EPSG:28992)
    category:
        Building category based on the SBR A table 10.1.
    monumental:
        Has the building structure a monumental status. Based on the SBR A table 10.3.
    structural_condition:
        Based on the SBR A table 10.2.
    vibration_sensitive:
        Has the building structure a vibration sensitive foundation. Based on the SBR A chapter 10.2.5.
    thickness:
        Layer thickness settlement-sensitive layer [m]
    foundation_element:
        Based on CUR 166 3rd edition table 5.19
    material_floor:
        Based on CUR 166 3rd edition table 5.20
    feature:
        default is bag:pand
        item in the BAG, see https://www.nationaalgeoregister.nl/geonetwork/srv/dut/catalog.search#/metadata/1c0dcc64-91aa-4d44-a9e3-54355556f5e7
    pagesize:
        Results per page

    Returns
    -------
    gdf: gpd.GeoDataFrame
    """

    if west > east:
        raise ValueError("west coordinate is larger than east coordinate")

    if south > north:
        raise ValueError("south coordinate is larger than north coordinate")

    if east - west > THRESHOLD:
        raise ValueError(
            f"x dimension of the bbox is {east - west} meters, larger than threshold"
        )

    if north - south > THRESHOLD:
        raise ValueError(
            f"y dimension of the bbox is {north - south} meters, larger than threshold"
        )

    wfs_query_params = {
        "service": "WFS",
        "version": "2.0.0",
        "request": "GetFeature",
        "typeName": feature,
        "srsname": "EPSG:28992",
        "outputFormat": "json",
        "bbox": str(west) + "," + str(south) + "," + str(east) + "," + str(north),
        "count": pagesize,
    }

    response = requests.get(
        url=BAG_WFS_URL,
        headers={"Content-Type": "application/json"},
        params=wfs_query_params,
        timeout=5,
    )
    if not response.ok:
        raise RuntimeError(response.text)

    gdf = gpd.read_file(json.dumps(response.json()), driver="GeoJSON").to_crs(
        "EPSG:28992"
    )

    # add default values
    gdf["name"] = gdf.index.astype(str)

    # sbr-A
    gdf["category"] = category
    gdf["structuralCondition"] = structural_condition
    gdf["vibrationSensitive"] = vibration_sensitive
    gdf["thickness"] = thickness
    gdf["monumental"] = monumental

    # prepal
    gdf["buildingDepth"] = np.clip(np.sqrt(gdf.area), 1.0, 18.0)
    gdf["buildingDepthVibrationSensitive"] = 1.0

    # cur166
    gdf["foundationElement"] = foundation_element
    gdf["material"] = material_floor

    return gdf


def create_prepal_payload(
    buildings: gpd.GeoDataFrame,
    location: Polygon | LineString | Point,
    pile_shape: Literal["square", "round"],
    pile_size: float,
    cone_resistance: float,
    reduction: float = 0.0,
    unit_weight: float = 20,
    elastic_modulus_factor: float = 15,
    poisson_ratio: float = 0.2,
    frequency: float = 20.0,
    vibration_type: Literal[
        "short-term", "repeated-short-term", "continuous"
    ] = "continuous",
    frequency_vibration_sensitive: float = 40.0,
    measurement_type: Literal["indicative", "limited", "extensive"] = "extensive",
    hysteretic_damping_barkan: float = -0.05,
) -> dict:
    """
    Create payload for VibraCore call `cur166/validation/multi`

    Parameters
    ----------
    buildings
        GeoDataFrame that holds teh building information. Must have the following columns:
            - buildingDepth:
              The minimum building depth [m]
            - buildingDepthVibrationSensitive:
              The minimum building depth [m]
            - category
              Based on the SBR A table 10.1.
            - monumental
              Has the building structure a monumental status. Based on the SBR A table 10.3.
            - structuralCondition
              Based on the SBR A table 10.2.
            - thickness
              Layer thickness settlement-sensitive layer [m]
            - vibrationSensitive
              Has the building structure a vibration sensitive foundation. Based on the SBR A chapter 10.2.5.
    location:
        Location of the source
    poisson_ratio:
        Poisson’s ratio of the soil [-]
    elastic_modulus_factor:
        Elastic modulus factor of the soil [-].
    unit_weight:
        Volume weight of the soil [kN/m^3]
    cone_resistance:
        Cone resistance [MPa]
    reduction:
        Reduction of the cone resistance [%]
    pile_size:
        Size of the pile [m]
    pile_shape:
        Shape of the pile.
    frequency:
        The dominate frequency [Hz]
    vibration_type
        Based on the SBR A table 10.4.
    frequency_vibration_sensitive
        The dominate frequency for vibration sensitive building [Hz].
    measurement_type
        Type of measurement based on the SBR A table 9.2.
    hysteretic_damping_barkan:
        hysteretic damping barkan [m^-1]

    Returns
    -------
    payload: dict

    Raises
    -------
    KeyError:
        Missing column names in GeoDataFrame
    """
    columns = [
        "category",
        "structuralCondition",
        "vibrationSensitive",
        "thickness",
        "buildingDepth",
    ]
    if not set(columns).issubset(set(buildings.columns)):
        msg = (
            f"Column names:{list(set(columns) - set(buildings.columns))} must be in GeoDataFrame. "
            f"Found column names: {buildings.columns}"
        )
        raise KeyError(msg)

    payload = {
        "buildingInformation": [
            {
                "geometry": mapping(row.geometry),
                "metadata": {"ID": row.get("name", uuid.uuid4().__str__())},
                "properties_PrePal": {
                    "buildingDepth": row["buildingDepth"],
                    "buildingDepthVibrationSensitive": row.get(
                        "buildingDepthVibrationSensitive", 1
                    ),
                    "calculationHeight": None,
                },
                "properties_SBRa": {
                    "category": row["category"],
                    "frequency": frequency,
                    "frequencyVibrationSensitive": frequency_vibration_sensitive,
                    "monumental": row["monumental"],
                    "structuralCondition": row["structuralCondition"],
                    "thickness": row["thickness"],
                    "vibrationSensitive": row["vibrationSensitive"],
                    "vibrationType": vibration_type,
                },
            }
            for i, row in buildings.iterrows()
        ],
        "vibrationSource": {"shape": pile_shape, "size": pile_size},
        "soilProperties": {
            "coneResistance": cone_resistance * (100 - reduction) / 100,
            "elasticModulus": cone_resistance * elastic_modulus_factor,
            "poissonRatio": poisson_ratio,
            "unitWeight": unit_weight,
        },
        "prediction": {
            "hystereticDampingBarkan": hysteretic_damping_barkan,
            "measurementType": measurement_type,
        },
        "validation": {"sourceLocation": mapping(location)},
    }

    return payload


def create_cur166_payload(
    buildings: gpd.GeoDataFrame,
    location: Polygon | LineString | Point,
    force: float,
    reduction: float = 0,
    installation_type: Literal["vibrate", "driving"] = "vibrate",
    building_part: Literal["floor", "wall"] = "floor",
    safety_factor: float = 0.05,
    vibration_direction: Literal["vertical", "horizontal"] = "vertical",
    frequency: float = 30.0,
    vibration_type: Literal[
        "short-term", "repeated-short-term", "continuous"
    ] = "continuous",
    frequency_vibration_sensitive: float = 40.0,
    reference_location: Literal[
        "Amsterdam",
        "Maasvlakte",
        "Rotterdam",
        "Groningen",
        "Den Haag",
        "Tiel",
        "Eindhoven",
    ] = "Amsterdam",
    measurement_type: Literal["indicative", "limited", "extensive"] = "extensive",
    methode_safety_factor: Literal["CUR", "exact"] = "exact",
) -> dict:
    """
    Create payload for VibraCore call `cur166/validation/multi`

    Parameters
    ----------
    buildings
        GeoDataFrame that holds teh building information. Must have the following columns:
            - foundationElement
              Based on CUR 166 3rd edition table 5.19
            - material
              Based on CUR 166 3rd edition table 5.20
            - category
              Based on the SBR A table 10.1.
            - monumental
              Has the building structure a monumental status. Based on the SBR A table 10.3.
            - structuralCondition
              Based on the SBR A table 10.2.
            - thickness
              Layer thickness settlement-sensitive layer [m]
            - vibrationSensitive
              Has the building structure a vibration sensitive foundation. Based on the SBR A chapter 10.2.5.
    location:
        Location of the source
    force:
        Impact force of the pile [kN]
    reduction:
        Reduction of impact [%]
    installation_type
        Based on CUR 166 3rd edition table 5.20 or 5.21
    building_part
        Based on CUR 166 3rd edition table 5.20 or 5.21
    safety_factor
        Based on CUR 166 3rd edition table 5.22
    vibration_direction
        Based on CUR 166 3rd edition table 5.22
    frequency:
        The dominate frequency [Hz]
    vibration_type
        Based on the SBR A table 10.4.
    frequency_vibration_sensitive
        The dominate frequency for vibration sensitive building [Hz].
        If not provided the frequency of 20 Hz is used for PrePal and 40
        for CUR 166 3rd edition high frequency vibration calculation.
    reference_location
        Based on CUR 166-1997 table 5.16 and 5.17
    measurement_type
        Type of measurement based on the SBR A table 9.2.
    methode_safety_factor
        Parameter that indicated how the safety factor is calculated.
        Find more info about the exact method here -> https://issuu.com/uitgeverijeducom/docs/geo_okt2014_totaal_v4_klein/36

    Returns
    -------
    payload: dict

    Raises
    -------
    ValueError:
        No reference values found for reference location
    KeyError:
        Missing column names in GeoDataFrame
    """
    reference = next(
        (
            item
            for item in SOIL_REFERENCE
            if item["location"] == reference_location
            and item["method"] == installation_type
            and item["vibration_direction"] == vibration_direction
        ),
        None,
    )

    if reference is None:
        raise ValueError(
            f"No reference values found for reference location: {reference_location}"
            f"with installation type: {installation_type} and vibration direction: {vibration_direction}."
        )

    columns = [
        "foundationElement",
        "material",
        "category",
        "monumental",
        "structuralCondition",
        "thickness",
        "vibrationSensitive",
    ]
    if not set(columns).issubset(set(buildings.columns)):
        msg = (
            f"Column names:{list(set(columns) - set(buildings.columns))} must be in GeoDataFrame. "
            f"Found column names: {buildings.columns}"
        )
        raise KeyError(msg)

    payload = {
        "buildingInformation": [
            {
                "geometry": mapping(row.geometry),
                "metadata": {"ID": row.get("name", uuid.uuid4().__str__())},
                "properties_CUR": {
                    "buildingPart": building_part,
                    "foundationElement": row["foundationElement"],
                    "installationType": installation_type,
                    "material": row["material"],
                    "safetyFactor": safety_factor,
                    "vibrationDirection": vibration_direction,
                },
                "properties_SBRa": {
                    "category": row["category"],
                    "frequency": frequency,
                    "frequencyVibrationSensitive": frequency_vibration_sensitive,
                    "monumental": row["monumental"],
                    "structuralCondition": row["structuralCondition"],
                    "thickness": row["thickness"],
                    "vibrationSensitive": row["vibrationSensitive"],
                    "vibrationType": vibration_type,
                },
            }
            for _, row in buildings.iterrows()
        ],
        "prediction": {
            "hystereticDampingBarkan": reference["alpha"],
            "force": force * (100 - reduction) / 100,
            "measurementType": measurement_type,
            "methodeSafetyFactor": methode_safety_factor,
            "referencesVelocity": reference["Uo"],
            "variationCoefficient": reference["Vo"],
        },
        "validation": {"sourceLocation": mapping(location)},
    }

    return payload


def get_normative_building(
    buildings: gpd.GeoDataFrame,
    location: Polygon | LineString | Point,
    category: Literal["one", "two"],
) -> str | None:
    """
    Get the name of the closest building

    Parameters
    ----------
    buildings:
        GeoDataFrame that holds the building information
    location:
        Geometry of the source location
    category:
        building category based on the SBR A table 10.1.

    Returns
    -------
    name: str
    """
    gdf = buildings.get(buildings["category"] == category)
    if gdf.empty:
        logging.error(f"ValueError: No buildings with category {category}.")
        return None
    gdf["distance"] = gdf.distance(location)
    return gdf.sort_values("distance", na_position="last").iloc[0].get("name")


def create_single_payload(
    multi_vibration_payload: dict,
    name: str,
) -> dict:
    """
    Create payload for VibraCore call `cur166/validation/single` or `prepal/validation/single`

    Parameters
    ----------
    multi_vibration_payload:
        result from `create_cur166_payload` or `create_prepal_payload`
    name:
        building name

    Returns
    -------
    payload: dict
    """
    payload = deepcopy(multi_vibration_payload)

    props = next(
        (
            item
            for item in payload["buildingInformation"]
            if item["metadata"]["ID"] == name
        ),
        None,
    )
    if props is None:
        raise ValueError(f"{name} is not a valid building name.")
    payload.update(dict(buildingInformation=props))

    return payload


def create_vibration_report_payload(
    multi_vibration_payload: dict,
    project_name: str,
    project_id: str,
    author: str,
) -> dict:
    """
    Creates a dictionary with the payload content for the VibraCore endpoint
    "/cur166/report" or "/prepal/report"

    This dictionary can be passed directly to `nuclei.client.call_endpoint()`.

    Parameters
    ----------
    vibration_payload:
        The result of a call to `create_cur166_payload()` or `create_prepal_payload()`
    project_name:
        The name of the project.
    project_id:
        The identifier (code) of the project.
    author:
        The author of the report.

    Returns
    -------
    payload: dict
    """
    payload = deepcopy(multi_vibration_payload)
    payload.update(
        dict(
            reportProperties=dict(
                author=author,
                projectNumber=project_id,
                projectName=project_name,
            ),
        )
    )
    return payload



from __future__ import annotations

import logging
from typing import Any, List, Tuple

import geopandas as gpd
import matplotlib.patches as patches
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from numpy.typing import NDArray
from scipy import interpolate
from shapely.geometry import LineString, Point, Polygon

from pyvibracore.results.plot_utils import _north_arrow, _scalebar


def _sound_prediction(
    power: float, k2: float, period: float, levels: List[float]
) -> NDArray:
    """
    Based on the 'Handleiding meten en rekenen industrielawaai' 2004 methode l.
    More information: https://open.overheid.nl/Details/ronl-15eb5528-d835-4f6a-b3a1-fc0851b334f9/1

    Parameters
    ----------
    power:
        source power [dB]
    k2:
        Correction term [dB]
    period:
        Operating period of the building code [hours]
    levels: list
        Array of floats to calculate the distance

    Returns
    -------
    distance: NDArray
    """
    distance = np.arange(1e-5, 500, step=0.2)
    noise = (
        power
        - (-10 * np.log10(period / 12))
        - (20 * np.log10(distance) + 0.005 * distance + 9.1)
        + k2
    )

    # interpolate and predict
    f = interpolate.interp1d(
        noise, distance, kind="cubic", assume_sorted=False, fill_value="extrapolate"
    )
    space = f(levels)

    # raise warning
    if any([item > 500 for item in space]):
        logging.warning(
            "One or more distances exceeds the 500 meter mark. "
            "Please note that this methode extrapolate the values from this point."
        )

    return space


def map_sound(
    buildings: gpd.GeoDataFrame,
    source_location: Point | LineString | Polygon,
    building_name: str,
    power: float,
    k2: float,
    period: float,
    title: str = "Legend:",
    figsize: Tuple[float, float] = (10.0, 12.0),
    settings: dict | None = None,
    **kwargs: Any,
) -> plt.Figure:
    """
    Create map of the input building settings.

    Parameters
    ----------
    buildings:
        GeoDataFrame of the input buildings
    source_location:
        location of the vibration source
    building_name:
        name of the building
    power:
        source power [dB]
    k2:
        Correction term [dB]
    period:
        Operating period of the building code [hours]
    title:
        Legend title
    figsize:
        Size of the activate figure, as the `plt.figure()` argument.
    settings:
        Plot settings used in plot: default settings are:

        .. code-block:: python

            {
                "source_location": {"label": "Trillingsbron", "color": "blue"},
                "levels": [
                    {
                        "label": ">80 db [0 dagen]",
                        "level": 80,
                        "color": "darkred",
                    },
                    {
                        "label": ">75 db [5 dagen]",
                        "level": 75,
                        "color": "red",
                    },
                    {
                        "label": ">70 db [15 dagen]",
                        "level": 70,
                        "color": "orange",
                    },
                    {
                        "label": ">65 db [30 dagen]",
                        "level": 65,
                        "color": "darkgreen",
                    },
                    {
                        "label": ">60 db [50 dagen]",
                        "level": 60,
                        "color": "green",
                    },
                ],
            }
    **kwargs:
        All additional keyword arguments are passed to the `pyplot.subplots()` call.

    Returns
    -------
    Figure
    """
    if settings is None:
        settings = {
            "source_location": {"label": "Trillingsbron", "color": "blue"},
            "levels": [
                {
                    "label": ">80 db [0 dagen]",
                    "level": 80,
                    "color": "darkred",
                },
                {
                    "label": ">75 db [5 dagen]",
                    "level": 75,
                    "color": "red",
                },
                {
                    "label": ">70 db [15 dagen]",
                    "level": 70,
                    "color": "orange",
                },
                {
                    "label": ">65 db [30 dagen]",
                    "level": 65,
                    "color": "darkgreen",
                },
                {
                    "label": ">60 db [50 dagen]",
                    "level": 60,
                    "color": "lightgreen",
                },
            ],
        }

    kwargs_subplot = {
        "figsize": figsize,
        "tight_layout": True,
    }

    kwargs_subplot.update(kwargs)

    fig, axes = plt.subplots(**kwargs_subplot)

    gpd.GeoSeries(source_location).plot(
        ax=axes, color=settings["source_location"]["color"], alpha=1, zorder=1, aspect=1
    )

    building = buildings.get(buildings["name"] == building_name)
    if building.empty:
        raise ValueError(f"No buildings with name {building_name}.")

    building.where(buildings["name"] == building_name).plot(
        ax=axes, zorder=2, color="gray", aspect=1
    )
    buildings.where(buildings["name"] != building_name).plot(
        ax=axes, zorder=2, color="lightgray", aspect=1
    )

    # plot contour
    levels = [values["level"] for values in settings["levels"]]
    distances = _sound_prediction(power, k2, period, levels=levels)
    colors = [values["color"] for values in settings["levels"]]
    for distance, color in zip(distances, colors):
        gpd.GeoSeries(building.buffer(distance).exterior).plot(
            ax=axes, zorder=3, color=color, aspect=1
        )

    # plot name
    for idx, row in buildings.iterrows():
        x = row.geometry.centroid.xy[0][0]
        y = row.geometry.centroid.xy[1][0]

        axes.annotate(
            idx,
            xy=(x, y),
            horizontalalignment="center",
        )

    # add legend
    axes.legend(
        title=title,
        title_fontsize=18,
        fontsize=15,
        loc="lower right",
        handles=[
            patches.Patch(
                facecolor=value["color"],
                label=value["label"],
                alpha=0.9,
                linewidth=2,
                edgecolor="black",
            )
            for value in settings["levels"]
        ],
    )

    _north_arrow(axes)
    _scalebar(axes)

    return fig


def get_normative_building(
    buildings: gpd.GeoDataFrame,
    location: Polygon | LineString | Point,
) -> str | None:
    """
    Get the name of the closest building with one of the follwing category:

        - "woonfunctie",
        - "gezondheidsfunctie",
        - "onderwijsfunctie"

    Parameters
    ----------
    buildings:
        GeoDataFrame that holds the building information
    location:
        Geometry of the source location

    Returns
    -------
    name: str
    """

    category = ["woonfunctie", "gezondheidsfunctie", "onderwijsfunctie"]

    gdf = buildings.get(
        [
            any(item in category for item in row.split(",")) if row else False
            for row in buildings["gebruiksdoel"].to_list()
        ]
    )
    if gdf.empty:
        logging.error(f"ValueError: No buildings with category {category}.")
        return None
    # FIXME: SettingWithCopyWarning
    with pd.option_context("mode.chained_assignment", None):
        gdf["distance"] = gdf.distance(location)
    return gdf.sort_values("distance", na_position="last").iloc[0].get("name")



import geopandas as gpd
import pytest

from pyvibracore.input.vibration_properties import (
    BAG_WFS_URL,
    create_cur166_payload,
    create_single_payload,
    get_buildings_geodataframe,
    get_normative_building,
)


def test_get_buildings_geodataframe(requests_mock, mock_bag_response):
    requests_mock.get(BAG_WFS_URL, json=mock_bag_response)

    gdf = get_buildings_geodataframe(1, 2, 3, 4)

    assert isinstance(gdf, gpd.GeoDataFrame)


def test_create_cur166_payload(requests_mock, mock_bag_response, mock_source_location):
    requests_mock.get(BAG_WFS_URL, json=mock_bag_response)
    gdf = get_buildings_geodataframe(1, 2, 3, 4)

    create_cur166_payload(
        gdf,
        location=mock_source_location,
        force=500,
    )

    with pytest.raises(ValueError):
        create_cur166_payload(
            gdf, location=mock_source_location, force=500, reference_location="Utrecht"
        )


def test_get_normative_building(requests_mock, mock_bag_response, mock_source_location):
    requests_mock.get(BAG_WFS_URL, json=mock_bag_response)
    gdf = get_buildings_geodataframe(1, 2, 3, 4)
    name = get_normative_building(
        gdf,
        location=mock_source_location,
        category="two",
    )

    assert isinstance(name, str)


def test_create_single_payload(requests_mock, mock_bag_response, mock_source_location):
    requests_mock.get(BAG_WFS_URL, json=mock_bag_response)
    gdf = get_buildings_geodataframe(1, 2, 3, 4)

    payload = create_cur166_payload(
        gdf,
        location=mock_source_location,
        force=500,
    )

    create_single_payload(payload, name="0")

    with pytest.raises(ValueError):
        create_single_payload(payload, name="-1")



import matplotlib.pyplot as plt
import pytest

from pyvibracore.input.impact_force_properties import (
    VibrationSource,
    create_multi_cpt_impact_force_payload,
    create_multi_cpt_impact_force_report_payload,
    create_single_cpt_impact_force_payload,
)
from pyvibracore.results.impact_force_result import MultiCalculationData


def test_vibration_source():
    VibrationSource.from_sheet_pile_name("AZ12-770")

    with pytest.raises(ValueError):
        VibrationSource.from_sheet_pile_name("abc")


def test_create_multi_cpt_impact_force_payload(cpt, mock_classify_response):
    create_multi_cpt_impact_force_payload(
        [cpt],
        {"S-TUN-016-PG": mock_classify_response},
        VibrationSource.from_sheet_pile_name("AZ12-770"),
        friction_strategy="CPTFrictionStrategy",
        drive_strategy="vibrate",
        installation_level_offset=-20,
    )


def test_create_multi_cpt_impact_force_report_payload(cpt, mock_classify_response):
    payload = create_multi_cpt_impact_force_payload(
        [cpt],
        {"S-TUN-016-PG": mock_classify_response},
        VibrationSource.from_sheet_pile_name("AZ12-770"),
        friction_strategy="CPTFrictionStrategy",
        drive_strategy="vibrate",
        installation_level_offset=-20,
    )
    create_multi_cpt_impact_force_report_payload(
        payload,
        project_name="Test project",
        project_id="123",
        author="Test User",
    )


def test_create_single_cpt_impact_force_payload(cpt, mock_classify_response):
    payload = create_multi_cpt_impact_force_payload(
        [cpt],
        {"S-TUN-016-PG": mock_classify_response},
        VibrationSource.from_sheet_pile_name("AZ12-770"),
        friction_strategy="CPTFrictionStrategy",
        drive_strategy="vibrate",
        installation_level_offset=-20,
    )

    create_single_cpt_impact_force_payload(payload, "S-TUN-016-PG")


def test_multi_calculation_data(mock_impact_force_response):
    result = MultiCalculationData.from_api_response(mock_impact_force_response)

    assert isinstance(result.plot(), plt.Figure)
